/*
   Feet of Fury

   drawables/bubble.cc

   Copyright (C)2002 Cryptic Allusion, LLC

   See the file LICENSE for more info on usage.
*/

#include "global.h"
#include "bubble.h"

Bubble::Bubble(Bubble::ColorScheme color) {
	m_phase = 0.0f;
	m_radius = 1.0f;
	m_stacks = 20;
	m_slices = 20;
	m_color = color;
	m_throbber = new BubbleThrobber();
	animAdd(m_throbber);

	pvr_poly_cxt_col(&m_cxt, PVR_LIST_TR_POLY);
	m_cxt.gen.culling = PVR_CULLING_NONE;
	pvr_poly_compile(&m_hdr, &m_cxt);
}

Bubble::~Bubble() { }

void Bubble::draw(int list) {
	pushTransformMatrix();
	if (list == PLX_LIST_TR_POLY) {
		switch (m_color) {
		case BlueGreen:
			drawSphereBlue();
			break;
		case Gray:
			drawSphereGray();
			break;
		case RedGreen:
			drawSphereRed();
			break;
		}
	}
	Drawable::draw(list);
	popTransformMatrix();
}

void Bubble::nextFrame() {
	m_phase += 2.0*M_PI / 240.0f;
	Drawable::nextFrame();
}

/*
   Draws a sphere out of triangle strips
   (based very loosely on a routine from James Susinno)
*/
void Bubble::drawSphereBlue() {
	int i, j;
	float alpha = getColor().a * 0.5f;
	pvr_dr_state_t dr_state;

	/* Setup our Direct Render state */
	pvr_dr_init(dr_state);

	/* Initialize xmtrx with the values from Parallax */
	plx_mat3d_apply_all();

	/* Put our own polygon header */
	pvr_prim(&m_hdr, sizeof(m_hdr));

	/* Iterate over stacks */
	for ( i=0; i<m_stacks; i++ ) {
		float pitch, pitch2;
		pitch = 2*M_PI * ( (float)i/(float)m_stacks );
		pitch2 = 2*M_PI * ( (float)(i+1)/(float)m_stacks );

		/* Iterate over slices: each entire stack will be one
		   long triangle strip. */
		for ( j=0; j<=m_slices/2; j++ ) {
			float x, y, z, g, b;
			float yaw, yaw2;
            
			yaw = 2*M_PI * ( (float)j/(float)m_slices );
			yaw2 = 2*M_PI * ( (float)(j+1)/(float)m_slices );

			/* x, y+1 */
			x = m_radius * fcos( yaw ) * fcos( pitch2 );
			y = m_radius * fsin( pitch2 );
			z = m_radius * fsin( yaw ) * fcos( pitch2 );
			mat_trans_single(x, y, z);
			g = fcos( yaw*2 ) / 2.0f + 0.5f;
			b = fsin( m_phase + pitch2 ) / 2.0f + 0.5f;
			plx_vert_fnd(&dr_state, PVR_CMD_VERTEX,
				x, y, z < 0.0001 ? 0.0001 : z,
				alpha, 0.0f, g, b);
                
			/* x, y */
			x = m_radius * fcos( yaw ) * fcos( pitch );
			y = m_radius * fsin( pitch );
			z = m_radius * fsin( yaw ) * fcos( pitch );
			mat_trans_single(x, y, z);
			g = fcos( yaw*2 ) / 2.0f + 0.5f;
			b = fsin( m_phase + pitch ) / 2.0f + 0.5f;
			if (j == (m_slices/2)) {
				plx_vert_fnd(&dr_state, PVR_CMD_VERTEX_EOL,
					x, y, z < 0.0001 ? 0.0001 : z,
					alpha, 0.0f, g, b);
			} else {
				plx_vert_fnd(&dr_state, PVR_CMD_VERTEX,
					x, y, z < 0.0001 ? 0.0001 : z,
					alpha, 0.0f, g, b);
			}
		}
	}
}

// Evil cut and paste copy (to maintain the DR speed) for gray spheres
void Bubble::drawSphereGray() {
	int i, j;
	uint32 alpha;
	pvr_dr_state_t dr_state;

	/* Calucate the alpha mask */
	alpha = ((int)(getColor().a * 0.5f * 255)) << 24;

	/* Setup our Direct Render state */
	pvr_dr_init(dr_state);

	/* Initialize xmtrx with the values from KGL */
	plx_mat3d_apply_all();

	/* Put our own polygon header */
	pvr_prim(&m_hdr, sizeof(m_hdr));

	/* Iterate over stacks */
	for ( i=0; i<m_stacks; i++ ) {
		float pitch, pitch2;
		pitch = 2*M_PI * ( (float)i/(float)m_stacks );
		pitch2 = 2*M_PI * ( (float)(i+1)/(float)m_stacks );

		/* Iterate over slices: each entire stack will be one
		   long triangle strip. */
		for ( j=0; j<=m_slices/2; j++ ) {
			float x, y, z, g;
			float yaw, yaw2;
			pvr_vertex_t *vert;
            
			yaw = 2*M_PI * ( (float)j/(float)m_slices );
			yaw2 = 2*M_PI * ( (float)(j+1)/(float)m_slices );

			/* x, y+1 */
			vert = pvr_dr_target(dr_state);
			vert->flags = PVR_CMD_VERTEX;
			x = m_radius * fcos( yaw ) * fcos( pitch2 );
			y = m_radius * fsin( pitch2 );
			z = m_radius * fsin( yaw ) * fcos( pitch2 );
			mat_trans_single(x, y, z);
			vert->x = x;
			vert->y = y;
			vert->z = z < 0.0001 ? 0.0001 : z;
			vert->u = 0.0f;
			vert->v = 0.0f;
			g = fcos( yaw*2 ) / 4.0f + 0.25f;
			// g = fsin( m_phase + pitch2 ) / 2.0f + 0.5f;
			vert->argb = alpha
				| ((uint32)(g * 255) << 16)
				| ((uint32)(g * 255) << 8)
				| ((uint32)(g * 255) << 0);
			vert->oargb = 0x00000000;
			pvr_dr_commit(vert);
                
			/* x, y */
			vert = pvr_dr_target(dr_state);
			if (j == (m_slices/2))
				vert->flags = PVR_CMD_VERTEX_EOL;
			else
				vert->flags = PVR_CMD_VERTEX;
			x = m_radius * fcos( yaw ) * fcos( pitch );
			y = m_radius * fsin( pitch );
			z = m_radius * fsin( yaw ) * fcos( pitch );
			mat_trans_single(x, y, z);
			vert->x = x;
			vert->y = y;
			vert->z = z < 0.0001 ? 0.0001 : z;
			vert->u = 0.0f;
			vert->v = 0.0f;
			g = fcos( yaw*2 ) / 4.0f + 0.25f;
			// g = fsin( m_phase + pitch ) / 2.0f + 0.5f;
			vert->argb = alpha
				| ((uint32)(g * 255) << 16)
				| ((uint32)(g * 255) << 8)
				| ((uint32)(g * 255) << 0);
			vert->oargb = 0x00000000;
			pvr_dr_commit(vert);
		}
	}
}


// Yet another evil cut and paste copy for red/yellow spheres
void Bubble::drawSphereRed() {
	int i, j;
	uint32 alpha;
	pvr_dr_state_t dr_state;

	/* Calucate the alpha mask */
	alpha = ((int)(getColor().a * 0.5f * 255)) << 24;

	/* Setup our Direct Render state */
	pvr_dr_init(dr_state);

	/* Initialize xmtrx with the values from KGL */
	plx_mat3d_apply_all();

	/* Put our own polygon header */
	pvr_prim(&m_hdr, sizeof(m_hdr));

	/* Iterate over stacks */
	for ( i=0; i<m_stacks; i++ ) {
		float pitch, pitch2;
		pitch = 2*M_PI * ( (float)i/(float)m_stacks );
		pitch2 = 2*M_PI * ( (float)(i+1)/(float)m_stacks );

		/* Iterate over slices: each entire stack will be one
		   long triangle strip. */
		for ( j=0; j<=m_slices/2; j++ ) {
			float x, y, z, g, b;
			float yaw, yaw2;
			pvr_vertex_t *vert;
            
			yaw = 2*M_PI * ( (float)j/(float)m_slices );
			yaw2 = 2*M_PI * ( (float)(j+1)/(float)m_slices );

			/* x, y+1 */
			vert = pvr_dr_target(dr_state);
			vert->flags = PVR_CMD_VERTEX;
			x = m_radius * fcos( yaw ) * fcos( pitch2 );
			y = m_radius * fsin( pitch2 );
			z = m_radius * fsin( yaw ) * fcos( pitch2 );
			mat_trans_single(x, y, z);
			vert->x = x;
			vert->y = y;
			vert->z = z < 0.0001 ? 0.0001 : z;
			vert->u = 0.0f;
			vert->v = 0.0f;
			g = fcos( yaw*2 ) / 2.0f + 0.5f;
			b = fsin( m_phase + pitch2 ) / 2.0f + 0.5f;
			vert->argb = alpha
				| ((uint32)(g * 255) << 16)
				| ((uint32)(b * 255) << 8);
			vert->oargb = 0x00000000;
			pvr_dr_commit(vert);
                
			/* x, y */
			vert = pvr_dr_target(dr_state);
			if (j == (m_slices/2))
				vert->flags = PVR_CMD_VERTEX_EOL;
			else
				vert->flags = PVR_CMD_VERTEX;
			x = m_radius * fcos( yaw ) * fcos( pitch );
			y = m_radius * fsin( pitch );
			z = m_radius * fsin( yaw ) * fcos( pitch );
			mat_trans_single(x, y, z);
			vert->x = x;
			vert->y = y;
			vert->z = z < 0.0001 ? 0.0001 : z;
			vert->u = 0.0f;
			vert->v = 0.0f;
			g = fcos( yaw*2 ) / 2.0f + 0.5f;
			b = fsin( m_phase + pitch2 ) / 2.0f + 0.5f;
			vert->argb = alpha
				| ((uint32)(g * 255) << 16)
				| ((uint32)(b * 255) << 8);
			vert->oargb = 0x00000000;
			pvr_dr_commit(vert);
		}
	}
}


