/*
   Feet of Fury

   drawables/plasma.cc

   Copyright (C)2002 Cryptic Allusion, LLC

   See the file LICENSE for more info on usage.
*/

#include "global.h"
#include "plasma.h"
#include "hsv.h"

/*
   Simple plasma effect from Stars 

   This will be used as the background texture for the bubbles effect.
   We use an especially nasty trick here to make the PVR do the dirty
   work for what used to be an impossibly slow task. Basically the
   plasma generator only generates a tiny 64x64 window of plasma, but
   into a twiddled texture. The PVR then uses this texture with
   bi-linear filtering to produce a nice smooth background.

   As far as I know, this is one of the few real "demo" effects I've
   seen for the DC so far, meaning a trick that makes the hardware do
   something it wasn't really meant to do. There are probably others
   that people have done but it's hard to tell without the source to
   a lot of it.
 */

/* Linear/iterative twiddling algorithm from Marcus' tatest */
#define TWIDTAB(x) ( (x&1)|((x&2)<<1)|((x&4)<<2)|((x&8)<<3)|((x&16)<<4)| \
        ((x&32)<<5)|((x&64)<<6)|((x&128)<<7)|((x&256)<<8)|((x&512)<<9) )
#define TWIDOUT(x, y) ( TWIDTAB((y)) | (TWIDTAB((x)) << 1) )
        
static short pcos[512], psin[512];
static int initted = 0;
static void plasma_init() {
	if (initted)
		return;
	for (int i=0; i<512; i++) {
		pcos[i] = (short)(fcos(i*2*(2*M_PI/512.0f)) * 256.0f);
		psin[i] = (short)(fsin(i*2*(2*M_PI/512.0f)) * 256.0f);
	}
	initted = 1;
}

Plasma::Plasma(int list) {
	m_list = list;
	m_xang = m_yang = 0;
	m_xamp = m_yamp = 256;
	m_hsv = 0.0f;

	m_txr[0] = pvr_mem_malloc(64*64*2);
	m_txr[1] = pvr_mem_malloc(64*64*2);
	m_w = m_h = 64.0f;

	pvr_poly_cxt_t cxt;
	pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565,
		64, 64, m_txr[0], PVR_FILTER_BILINEAR);
	pvr_poly_compile(&m_hdr[0], &cxt);
	pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565,
		64, 64, m_txr[1], PVR_FILTER_BILINEAR);
	pvr_poly_compile(&m_hdr[1], &cxt);

	m_txr_cur = 0;
	
	plasma_init();
}

Plasma::~Plasma() {
	pvr_mem_free(m_txr[0]);
	pvr_mem_free(m_txr[1]);
}

void Plasma::setSize(float w, float h) {
	m_w = w;
	m_h = h;
}

void Plasma::draw(int list) {
	if (list != m_list)
		return;

	Vector p = getPosition();
	Color c = getColor();
	
	float r, g, b;
	hsvSpectrum(m_hsv, r, g, b);
	r *= c.r; g *= c.g; b *= c.b;

	pvr_prim(&m_hdr[m_txr_cur], sizeof(pvr_poly_hdr_t));

	pvr_vertex_t	v;
	v.flags = PVR_CMD_VERTEX;
	v.x = p.x - m_w/2;
	v.y = p.y + m_h/2;
	v.z = p.z;
	v.u = 0.0f;
	v.v = 1.0f;
	//v.argb = PVR_PACK_COLOR(1.0f, 2*55/256.0f, 2*77/256.0f, 2*142/256.0f);
	v.argb = PVR_PACK_COLOR(1.0f, r, g, b);
	v.oargb = 0;
	pvr_prim(&v, sizeof(v));

	v.y = p.y - m_h/2;
	v.v = 0.0f;
	pvr_prim(&v, sizeof(v));

	v.x = p.x + m_w/2;
	v.y = p.y + m_h/2;
	v.u = 1.0f; v.v = 1.0f;
	pvr_prim(&v, sizeof(v));

	v.flags = PVR_CMD_VERTEX_EOL;
	v.y = p.y - m_h/2;
	v.v = 0.0f;
	pvr_prim(&v, sizeof(v));
}

void Plasma::nextFrame() {
	uint16	*vrout = (uint16 *)m_txr[m_txr_cur];
	int	x,y,p,q, r;

	m_xang++; m_yang--;
	m_xang &= 511; m_yang &= 511;
	m_xamp = 2*pcos[m_yang];
	m_yamp = 2*psin[m_xang];

	// Redo the plasma texture for the next frame
	for (y=30;y<30+64;y++) {
		for (x=30;x<30+64;x++) {
			p=x+(psin[(m_xang+(y*2+y))&511]*m_xamp/256);
			q=y+(pcos[(m_yang+(x*4  ))&511]*m_yamp/256);
			r=((p+q)&255);
			if (r >= 128)
				r = 255 - r;
			p = (r >> (8 - 5));
			r = (r >> (8 - 6));
			vrout[TWIDOUT(x - 30, y - 30)] =
				(p << 11) | (r << 5) | p;
		}
	}

	m_hsv += 1.0f / (8*hz);
	while (m_hsv >= 1.0f)
		m_hsv -= 1.0f;

	m_txr_cur ^= 1;
}

