/*
   Feet of Fury

   bsod.cc

   Copyright (C)2002 Cryptic Allusion, LLC

   See the file LICENSE for more info on usage.
*/

#include "global.h"
#include "bsod.h"

/* This module sets up handlers for all the crash points of the KOS
   kernel so that we can display a "nice" crash screen for the user
   to give us feed back from (as opposed to a black screen and an abrupt
   exit to the main menu). */

static void setup() {
	/* Make sure all the IRQs are disabled, just in case */
	irq_disable();

	/* Do some raw hardware twiddling, because we don't know how much
	   of the system is still functional... */
	vid_set_start(0);
	vid_clear(64, 0, 96);
}

static void finish() {
	for( ; ; )
		;
}

static char buffer[256];
static void irqHandler(irq_t source, irq_context_t * context) {
	arch_menu();

	int x = 40, y = 40;
	setup();
	
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Unhandled Exception (software failure)"); y += 24;

	sprintf(buffer, "Source %08lx", source);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;
	
	y += 24;

	sprintf(buffer, "R0-R3: %08lx %08lx %08lx %08lx",
		context->r[0], context->r[1], context->r[2], context->r[3]);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;
	sprintf(buffer, "R4-R7: %08lx %08lx %08lx %08lx",
		context->r[4], context->r[5], context->r[6], context->r[7]);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;
	sprintf(buffer, "R8-RB: %08lx %08lx %08lx %08lx",
		context->r[8], context->r[9], context->r[10], context->r[11]);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;
	sprintf(buffer, "RC-RF: %08lx %08lx %08lx %08lx",
		context->r[12], context->r[13], context->r[14], context->r[15]);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;

	sprintf(buffer, "PC: %08lx SR: %08lx PR: %08lx",
		context->pc, context->sr, context->pr);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;

	y += 24;

	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Please report to:"); y += 24;
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "  dpotter@cagames.com");
	
	y += 24;

	bfont_draw_str(vram_s + 640*y+x, 640, 0, "GAME HALTED");
	
	finish();
}

static void assertHandler(const char * file, int line, const char * expr,
	const char * msg, const char * func)
{
	arch_menu();

	uint32 fp = arch_get_fptr();
	int x = 40, y = 40;
	setup();

	if (expr == NULL)
		expr = "(none)";
	if (msg == NULL)
		msg = "(none)";
	if (func == NULL)
		func = "(none)";

	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Assertion Failure (software failure)"); y += 24;

	y += 24;

	sprintf(buffer, "Source %s:%d (%s)", file, line, func);
	bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Expression:"); y += 24;
	bfont_draw_str(vram_s + 640*y+x+5, 640, 0, (char *)expr); y += 24;
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Message:"); y += 24;
	bfont_draw_str(vram_s + 640*y+x+5, 640, 0, (char *)msg); y += 24;
	
	y += 24;

#ifdef FRAME_POINTERS
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Stack Trace:"); y += 24;
	while (fp != 0xffffffff && y < 300) {
		if ((fp & 3) || (fp < 0x8c000000) || (fp > 0x8d000000)) {
			bfont_draw_str(vram_s + 640*y+x, 640, 0, "(invalid fp)"); y += 24;
			break;
		}
		sprintf(buffer, "  %08lx", arch_fptr_ret_addr(fp));
		bfont_draw_str(vram_s + 640*y+x, 640, 0, buffer); y += 24;

		fp = arch_fptr_next(fp);
	}
#endif

	y += 24;

	bfont_draw_str(vram_s + 640*y+x, 640, 0, "Please report to:"); y += 24;
	bfont_draw_str(vram_s + 640*y+x, 640, 0, "  dpotter@cagames.com"); y += 24;

	y += 24;

	bfont_draw_str(vram_s + 640*y+x, 640, 0, "GAME HALTED");
	
	finish();
}

void setupBsod() {
	// Interrupt handlers
	irq_set_handler(EXC_UNHANDLED_EXC, irqHandler);
	irq_set_handler(EXC_DOUBLE_FAULT, irqHandler);

	// Assert handler
	assert_set_handler(assertHandler);
}

