/* 
   Feet of Fury
   
   main/loading.cc
   
   Copyright (C)2002 Cryptic Allusion, LLC

   See the file LICENSE for more info on usage.
*/

#include "global.h"
#include "loading.h"
#include "bubble.h"
#include "cube.h"

/*

Each instance of FoF will contain exactly one of these objects, and it is
managed by the loading_*() functions at the bottom.

This makes use of KOS threads to give the user something to look at while
a new part of the game is loading (e.g., transitioning from menu to stage
select, stage select to game play, etc). It assumes that when you call
loading_start(), it is free to take over the PVR rendering system and
do its own thing in a thread. When you call loading_finish(), it will
finish up whatever animation it's working on and relinquish control again.

*/

static void thdproc(void *ptr) {
	((Loading *)ptr)->threadproc();
}

Loading::Loading() {
	// Allocate a semaphore for our usage
	m_sem = sem_create(0);

	// Set our state to waiting for birth
	m_state = StateWaitBirth;

	// Init other vars
	m_bubble = NULL; m_cube = NULL;

	// Load our cube texture
	m_txr = new Texture("preload/gfx_misc/scrolly_logo.png", false);
	m_txrLoading = new Texture("preload/gfx_misc/loading.png", true);
	m_txrLoading->setUVClamp(Texture::UVClamp, Texture::UVClamp);

	// Start up the worker thread
	if (!thd_create(thdproc, (void *)this)) {
		assert_msg( false, "Can't create the Loading threadproc" );
	} else {
		// Wait for the threadproc to be ready
		while (m_state != StateReady)
			thd_pass();
	}
}

Loading::~Loading() {
	// Tell the thread to quit
	m_state = StateWaitDeath;
	signal();
	while (m_state != StateDead)
		thd_pass();

	// Delete our semaphore
	sem_destroy(m_sem);
}

// Do a frame of display
void Loading::doFrame() {
	if (m_state == StateRunning) {
		if (m_alpha < 1.0f)
			m_alpha += 0.05f;
	} else if (m_state == StateStopRunning) {
		if (m_alpha > 0.0f)
			m_alpha -= 0.1f;
	}
	if (m_alpha < 0.0f) m_alpha = 0.0f;
	if (m_alpha > 1.0f) m_alpha = 1.0f;

	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
	pvr_list_begin(PVR_LIST_TR_POLY);

	pvr_set_bg_color(0.0f, 0.0f, 0.0f);

	plx_cxt_culling(PLX_CULL_NONE);
	plx_cxt_texture(NULL);
	drawOuterRing();
	drawMidRing();
	drawInnerRing();
	drawLoadingRing();

	drawThrobber();

	pvr_scene_finish();

	m_bubble->nextFrame();
	m_frame++;
}

// Methods to draw the various rings
void Loading::drawOuterRing() {
	// Do the rotations
	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -10.0f);
	plx_mat3d_rotate(m_frame * (60.0f/hz) * 2, 0.0f, 1.0f, -0.5f);
	plx_mat3d_rotate(m_frame * (60.0f/hz) * 2.2f, 0.0f, 0.0f, -1.0f);
	plx_mat3d_apply_all();

	// Draw the ring
	const int segcnt = 120;
	const float rad = 6.0f;
	int i;
	uint32 color = plx_pack_color(m_alpha, 0.05f, 0.4f, 0.58f);
	plx_dr_state_t dr;

	plx_dr_init(&dr);
	plx_cxt_texture(NULL);
	plx_cxt_send(PLX_LIST_TR_POLY);

	for (i=0; i<=segcnt; i++) {
		// Get the x,y for the current outer point
		float t = i * ((2*M_PI)/segcnt);
		float cox = fcos(t) * rad;
		float coy = fsin(t) * rad;

		// Current inner point
		float cix = fcos(t) * rad * 0.9f;
		float ciy = fsin(t) * rad * 0.9f;

		// Adjust for peaks around the corners
		if ( (i % (segcnt/8)) == 0 ) {
			cix *= 0.9f;
			ciy *= 0.9f;
		}

		// Do we need to start a new strip?
		if (i>0 && i<segcnt && (i % 40) == 0) {
			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT_EOS, cix, ciy, 0.0f, color);

			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT, cix, ciy, 0.0f, color);
		} else if (i == segcnt) {
			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT_EOS, cix, ciy, 0.0f, color);
		} else {
			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT, cix, ciy, 0.0f, color);
		}
	}
}

void Loading::drawMidRing() {
	// Do the rotations
	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -10.0f);
	plx_mat3d_rotate(m_frame * (60.0f/hz) * 2, 1.0f, 0.0f, -0.9f);
	plx_mat3d_apply_all();

	// Draw the ring
	const int segcnt = 120;
	const float rad = 4.0f;
	int i;
	uint32 color = plx_pack_color(m_alpha, 0.09f, 0.56f, 0.79f);
	plx_dr_state_t dr;

	plx_dr_init(&dr);
	plx_cxt_texture(NULL);
	plx_cxt_send(PLX_LIST_TR_POLY);

	for (i=0; i<=segcnt; i++) {
		// Get the x,y for the current outer point
		float t = i * ((2*M_PI)/segcnt);
		float cox = fcos(t) * rad;
		float coy = fsin(t) * rad;

		// Current inner point
		float cix = fcos(t) * rad * 0.9f;
		float ciy = fsin(t) * rad * 0.9f;

		// If we're at a corner, then skip some
		if (i == segcnt || ((i + segcnt/16) % (segcnt/8)) == 0 ) {
			// End this strip
			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT_EOS, cix, ciy, 0.0f, color);

			i += 4;
		} else {
			// Draw two points
			plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
			plx_vert_indm3(&dr, PLX_VERT, cix, ciy, 0.0f, color);
		}
	}
}

void Loading::drawInnerRing() {
	// Do the rotations
	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -10.0f);
	plx_mat3d_rotate(m_frame * (60.0f/hz) * 2, 1.0f, 1.0f, -3.0f);
	plx_mat3d_apply_all();

	// Draw the ring
	const int segcnt = 120;
	const float rad = 2.0f;
	int i;
	uint32 color = plx_pack_color(m_alpha, 0.07f, 0.57f, 0.33f);
	plx_dr_state_t dr;

	plx_dr_init(&dr);
	plx_cxt_texture(NULL);
	plx_cxt_send(PLX_LIST_TR_POLY);

	for (i=0; i<=segcnt; i++) {
		// Get the x,y for the current outer point
		float t = i * ((2*M_PI)/segcnt);
		float cox = fcos(t) * rad;
		float coy = fsin(t) * rad;

		// Current inner point
		float cix = fcos(t) * rad * 0.9f;
		float ciy = fsin(t) * rad * 0.9f;

		// Draw two points
		plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
		plx_vert_indm3(&dr, PLX_VERT, cix, ciy, 0.0f, color);

		// Get the next points
		t = (i+1) * ((2*M_PI)/segcnt);
		cox = fcos(t) * rad;
		coy = fsin(t) * rad;
		cix = fcos(t) * rad * 0.9f;
		ciy = fsin(t) * rad * 0.9f;

		// Draw those too
		plx_vert_indm3(&dr, PLX_VERT, cox, coy, 0.0f, color);
		plx_vert_indm3(&dr, PLX_VERT_EOS, cix, ciy, 0.0f, color);

		// Skip some space
		i += 2;
	}
}

void Loading::drawLoadingRing() {
	// Do the rotations
	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -10.0f);
	// plx_mat3d_rotate(m_frame * (60.0f/hz) * 2, -0.2f, 0.8f, 0.4f);
	plx_mat3d_rotate(180.0f, 0.0f, 0.0f, 1.0f);
	plx_mat3d_rotate(-90.0f, 1.0f, 0.0f, 0.0f);
	plx_mat3d_rotate(m_frame * (60.0f/hz) * 2, 0.0f, 0.0f, 1.0f);
	plx_mat3d_apply_all();

	// Draw the ring
	const int segcnt = 120;
	const float rad = 2.5f; // 1.75f;
	int i;
	// uint32 color = plx_pack_color(m_alpha, 0.3f, 0.8f, 0.9f);
	uint32 color = plx_pack_color(m_alpha, 0.75f, 0.75f, 0.75f);
	plx_dr_state_t dr;

	plx_dr_init(&dr);
	plx_cxt_texture(*m_txrLoading);
	plx_cxt_send(PLX_LIST_TR_POLY);

	float h = fsin(m_frame * 2*M_PI / hz) / 4.0f + 0.5f;

	for (i=0; i<=segcnt; i++) {
		// Get the x,y for the point
		float t = i * ((2*M_PI)/segcnt);
		float cx = fcos(t) * rad;
		float cy = fsin(t) * rad;

		float u = 1.0f - (i * 2.0f / segcnt);

		plx_vert_ifdm3(&dr, PLX_VERT, cx, cy, +h, color, u, 1.0f);
		plx_vert_ifdm3(&dr, i == segcnt ? PLX_VERT_EOS : PLX_VERT, cx, cy, -h, color, u, 0.0f);
	}
}

// Draw the "throbber" cube/sphere
void Loading::drawThrobber() {
	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -10.0f);
	// glTranslatef(0, 0, -4.0f + fsin(m_frame * 4 * 2 * M_PI / 360.0f) * 2.0f);
	//glRotatef(m_frame, 0.7f, 1.0f, 0.6f);
	//glTranslatef(4.0f, 0.0f, 4.0f);
	plx_mat3d_rotate(m_frame * (60.0f / hz) * 2, 0.8f, 1.0f, 0.9f);
	plx_mat3d_scale(1.5f, 1.5f, 1.5f);
	m_bubble->setTint(Color(m_alpha, 1, 1, 1));
	m_bubble->draw(PLX_LIST_TR_POLY);
}
        
        
// Our thread will run in here
void Loading::threadproc() {
	printf("Loading::threadproc: started\n");

	// Tell constructor we're ready
	m_state = StateReady;

	while (m_state != StateWaitDeath) {
		switch (m_state) {
		case StateReady:
			// Wait for a signal
			wait();
			break;
		case StateRunning:
			// Do a frame of display
			doFrame();
			break;
		case StateStopRunning:
			doFrame();
			if (m_alpha <= 0.0f) {
				m_alpha = 0.0f;
				doFrame();
				doFrame();
				doFrame();
				m_state = StateReady;
			}
			break;
		}
	}
	
	printf("Loading::threadproc: finished\n");
	m_state = StateDead;
}

// Tell us to begin our fun display
void Loading::start() {
	if (m_state != StateReady) {
		printf("Loading::start: warning -- start() called but we're not StateReady\n");
		return;
	}

	m_frame = 0;
	m_alpha = 0.0f;
	m_bubble = new Bubble(Bubble::BlueGreen);
	m_cube = new Cube(PLX_LIST_TR_POLY, m_txr);
	m_cube->setScale(Vector(0.5f, 0.5f, 0.5f));
	m_bubble->subAdd(m_cube);
	m_bubble->setTint(Color(0.0f, 1, 1, 1));
	m_state = StateRunning;
	signal();
}

// Tell us to stop our fun display
void Loading::finish() {
	if (m_state != StateRunning) {
		printf("Loading::finish: warning -- finish() called but we're not StateRunning\n");
		return;
	}

	m_state = StateStopRunning;
	signal();
	while (m_state != StateReady)
		thd_pass();

	assert( m_bubble != NULL );
	m_bubble = NULL; m_cube = NULL;
}

// Signal / wait on the semaphore
void Loading::signal() {
	sem_signal(m_sem);
}

void Loading::wait() {
	sem_wait(m_sem);
}

///////////////////////////////////////////////////////////////////////

// Our one Loading object
static Loading * loading = NULL;

// Call to initialize/shutdown the loading twiddler
void loadingInit() {
	// Create a Loading object
	assert( loading == NULL );
	loading = new Loading();
}

void loadingShutdown() {
	// Tell the Loading object to quit
	assert( loading != NULL );
	delete loading; loading = NULL;
}

// Call when you start loading
void loadingStart() {
	// Tell the Loading object to start its display
	printf("loadingStart called\n");
	assert( loading != NULL );
	loading->start();
}

// Call when you're done loading
void loadingFinish() {
	// Tell the Loading object to stop its display
	printf("loadingFinish called\n");
	assert( loading != NULL );
	loading->finish();
}

